#!/usr/bin/env python
from setuptools import setup

setup()

# https://packaging.python.org/guides/single-sourcing-package-version/
# https://packaging.python.org/guides/distributing-packages-using-setuptools
# http://blog.ionelmc.ro/2014/05/25/python-packaging/
